/* ---------------------------------------------------------------------------*
 * FMU which models the exponantial decay curve.
 * This FMU implements der(x) = u -x, x0 = 1.0 and y = -x.
 *
 * This FMU has the following properties:
 *  - It has one continuous state variable.
 *  - It has one input and one output.
 *
 * This file is based on the template FMU 'helloWorldME2' developed by
 * Christopher Brooks and Edward A. Lee.
 *
 * Authors: Thierry Nouidui.
 * ---------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <math.h>

// See fmiFunctions.h
#ifdef FMI_STATIC_OR_C_FILE  // FMI_STATIC_OR_C_FILE is a Ptolemy-specific extension.
#define FMI_FUNCTION_PREFIX myFirstFMU_
#endif

// Include fmu header files, typedefs and macros
#include "fmi2Functions.h"

// Data structure for an instance of this FMU. */
typedef struct {
    fmi2Real    *r;
    fmi2Boolean mustComputeOutputs;
    const fmi2CallbackFunctions* functions;
    fmi2String instanceName;
} ModelInstance;

// Value references.
#define vr_x 0
#define vr_der_x 1
#define vr_u 2
#define vr_y 3

// Number of variables
#define NVARS 4

// Globally unique ID used to make sure the XML file and the DLL match.
// See also guid in modelDescription.xml
#define MODEL_GUID "{c9c3ec64-b950-48b0-8a67-05d216defe39}"

// Instantiate the FMU
FMI2_Export fmi2Component fmi2Instantiate(fmi2String instanceName,
        fmi2Type   fmuType,
        fmi2String fmuGUID,
        fmi2String fmuResourceLocation,
        const fmi2CallbackFunctions* functions,
        fmi2Boolean                  visible,
        fmi2Boolean                  loggingOn) {

    ModelInstance* component;

    // Perform checks.
    if (!functions->logger)
        return NULL;
    if (!functions->allocateMemory || !functions->freeMemory){
        functions->logger(NULL, instanceName, fmi2Error, "error",
                "fmiInstantiateSlave: Missing callback function: freeMemory");
        return NULL;
    }
    if (!instanceName || strlen(instanceName)==0) {
        functions->logger(NULL, instanceName, fmi2Error, "error",
                "fmiInstantiateSlave: Missing instance name.");
        return NULL;
    }
    if (strcmp(fmuGUID, MODEL_GUID)) {
        functions->logger(NULL, instanceName, fmi2Error, "error",
                "fmiInstantiateSlave: Wrong GUID %s. Expected %s.", fmuGUID, MODEL_GUID);
        return NULL;
    }
    component = (ModelInstance *)functions->allocateMemory(1, sizeof(ModelInstance));
    component->r = functions->allocateMemory(NVARS, sizeof(fmi2Real));
    component->functions = functions;
    component->instanceName = instanceName;
    component->mustComputeOutputs = fmi2True;

	// Set the start values of variables that define a start value.
	component->r[vr_x] = 1.0;
	component->r[vr_u] = 0.0;
    return component;
}

// Initialize the FMU.
FMI2_Export fmi2Status fmi2EnterInitializationMode(fmi2Component c) {
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2ExitInitializationMode(fmi2Component c) {
    return fmi2OK;
}
 
// Set the continuous states of the FMU.
FMI2_Export fmi2Status fmi2SetContinuousStates(fmi2Component c, const fmi2Real x[],
        size_t nx) {
    // Model Exchange
    ModelInstance* component = (ModelInstance *) c;
    if (nx == 1){
        component->r[vr_x] = x[0];
        // re-initialize caching of variables
        // that depend on states.
        component->r[vr_y] = -component->r[vr_x];
    }
    else{
        component->functions->logger(NULL, component->instanceName, fmi2Error, "error",
                          "fmi2SetContinuousStates: Call failed because nx is not 1.");
        return fmi2Error;
    }
    return fmi2OK;
}

// Set the inputs of the FMU.
FMI2_Export fmi2Status fmi2SetReal(fmi2Component c, const fmi2ValueReference vr[], size_t nvr, const fmi2Real value[]){
    int i;
    ModelInstance* component = (ModelInstance *) c;
    if (nvr > NVARS){
        component->functions->logger(NULL, component->instanceName, fmi2Error, "error",
                          "fmi2SetReal: Too many real arguments are provided.");
        return fmi2Error;
    }
    // Set values.
    for (i = 0; i < nvr; i++) {
        component->r[vr[i]] = value[i];
    }
    // Set a flag that indicates that the outputs must be re-computed.
    component->mustComputeOutputs = fmi2True;
    return fmi2OK;
}

// Get the derivatives of the FMU.
FMI2_Export fmi2Status fmi2GetDerivatives(fmi2Component c, fmi2Real derivatives[],
        size_t nx) {
    ModelInstance* component = (ModelInstance *) c;
    if (nx == 1){
      // If outputs are not current, then just evaluate the derivative
      // as this is all that is needed in this call.
      if (component->mustComputeOutputs)
        derivatives[0] = component->r[vr_u] - component->r[vr_x] ; 
      else
        derivatives[0] = component->r[vr_der_x];
    }
    else{
        component->functions->logger(NULL, component->instanceName, fmi2Error, "error",
                          "fmi2GetDerivatives: Call failed because nx is not 0 or 1.");
        return fmi2Error;
    }
    return fmi2OK;
}

// Get the outputs of the FMU
FMI2_Export fmi2Status fmi2GetReal(fmi2Component c,
                                const fmi2ValueReference vr[],
                                size_t nvr,
                                fmi2Real value[]) {
    int i;
    double dp;
    ModelInstance* component = (ModelInstance *) c;

    if (nvr > NVARS) {
        return fmi2Error;
    }
    if (nvr > 0) {
      // Check if the output must be computed.
      if (component->mustComputeOutputs){
        component->r[vr_y] = -component->r[vr_x];
        component->mustComputeOutputs = fmi2False;
      }
      // Assign outputs
      for(i=0; i < nvr; i++){
        value[i] = component->r[vr[i]];
      }
    }
    return fmi2OK;
    }


// Get the continuous states of the FMU.
FMI2_Export fmi2Status fmi2GetContinuousStates(fmi2Component c, fmi2Real x[],
        size_t nx) {
    ModelInstance* component = (ModelInstance *) c;
    if (nx == 1){
        x[0] = component->r[vr_x];
        return fmi2OK;
    }
    else{
        printf("myFirstFMU.c: Called fmi2GetContinuousStates with nx=%d.\n", (int)nx);
        component->functions->logger(NULL, component->instanceName, fmi2Error, "error",
                          "fmi2GetContinuousStates: Call failed because nx is not 1.");
        return fmi2Error;
    }
}

// Terminate the FMU.
FMI2_Export fmi2Status fmi2Terminate(fmi2Component c) {
    return fmi2OK;
}

// Free the FMU.
void fmi2FreeInstance(fmi2Component c) {
    ModelInstance* component = (ModelInstance *) c;
    component->functions->freeMemory(component->r);
    component->functions->freeMemory(component);
}


////////////////////////////////////////////////////////////
// Below here are functions that need to be in the FMU. 
// They are not used in this example but are needed by certain
// tools to check the validity of the FMU.

FMI2_Export fmi2Status fmi2Reset(fmi2Component c) {
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2GetInteger(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, fmi2Integer value[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2GetBoolean(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, fmi2Boolean value[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2GetString(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, fmi2String value[]) {
    return fmi2Error;
}

FMI2_Export const char* fmiGetTypesPlatform() {
    // We return a string literal, which does not require malloc.
    // Note that this is declared const char * because it is not safe to
    // modify a string literal in C.
    return "default";
}

FMI2_Export const char* fmi2GetVersion() {
    // We return a string literal, which does not require malloc.
    // Note that this is declared const char * because it is not safe to
    // modify a string literal in C.
    return "2.0";
}

const char* fmi2GetTypesPlatform() {
    return fmi2TypesPlatform;
}

FMI2_Export fmi2Status fmi2SetDebugLogging(fmi2Component c,
        fmi2Boolean      loggingOn,
        size_t          nCategories,
        const fmi2String categories[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SetInteger(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, const fmi2Integer value[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SetBoolean(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, const fmi2Boolean value[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SetString(fmi2Component c, const fmi2ValueReference vr[],
        size_t nvr, const fmi2String value[]) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SetupExperiment(fmi2Component c,
        fmi2Boolean   toleranceDefined,
        fmi2Real      tolerance,
        fmi2Real      startTime,
        fmi2Boolean   stopTimeDefined,
        fmi2Real      stopTime) {
    // There is nothing to do here.
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2GetFMUstate(fmi2Component c, fmi2FMUstate* FMUstate) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SetFMUstate(fmi2Component c, fmi2FMUstate FMUstate) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2FreeFMUstate(fmi2Component c, fmi2FMUstate* FMUstate) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SerializedFMUstateSize(fmi2Component c, fmi2FMUstate FMUstate,
        size_t* size) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2SerializedFMUstate(fmi2Component c, fmi2FMUstate FMUstate,
        fmi2Byte serializedState[], size_t size) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2DeSerializedFMUstate(fmi2Component c,
        const fmi2Byte serializedState[],
        size_t size, fmi2FMUstate* FMUstate) {
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2GetDirectionalDerivative(fmi2Component c,
        const fmi2ValueReference vUnknown_ref[], size_t nUnknown,
        const fmi2ValueReference vKnown_ref[],   size_t nKnown,
        const fmi2Real dvKnown[], fmi2Real dvUnknown[]) {
    printf("myFirstFMU.c: fmi2GetDirectionalDerivative() called, even though the FMU does not provide them.\n");
    // The standard 2.0, RC 1 says on p. 26:
    // If the capability attribute “providesDirectionalDerivative” is true,
    // fmi2GetDirectionalDerivative computes a linear combination of the partial derivatives of h
    // with respect to the selected input variables
    return fmi2Error;
}

// Start of Model Exchange functions.
// Alphabetical in this section.

FMI2_Export fmi2Status fmi2GetNominalsOfContinuousStates(fmi2Component c,
        fmi2Real x_nominal[],
        size_t nx) {
    int i;
    // The standard says to return 1.0 if the FMU has no information about
    // the nominal values
    for(i=0; i < nx; i++)
        x_nominal[i] = 1.0;
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2CompletedIntegratorStep(fmi2Component c,
        fmi2Boolean   noSetFMUStatePriorToCurrentPoint,
        fmi2Boolean*  enterEventMode,
        fmi2Boolean*   terminateSimulation) {
    // Model Exchange
    // Return enterEventMode = fmi2False to indicate that this FMU
    // stays in the continuous-time mode.
    *enterEventMode = fmi2False;
    *terminateSimulation = fmi2False;
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2EnterContinuousTimeMode(fmi2Component c) {
    // Model Exchange
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2EnterEventMode(fmi2Component c) {
    // Model Exchange
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2GetEventIndicators(fmi2Component c,
        fmi2Real eventIndicators[], size_t ni) {
    // Model Exchange
    printf("myFirstFMU.c: fmi2GetEventIndicators() returning fmi2Error as it does not trigger events.\n");
    return fmi2Error;
}

FMI2_Export fmi2Status fmi2NewDiscreteStates(fmi2Component  c,
        fmi2EventInfo* fmi2EventInfo) {
    // Model Exchange
    fmi2EventInfo->newDiscreteStatesNeeded = fmi2False;
    return fmi2OK;
}

FMI2_Export fmi2Status fmi2SetTime(fmi2Component c, fmi2Real time) {
    // Model Exchange
    return fmi2OK;
}



